# job_app.pyw
#
# Standalone MS Windows PyQGIS script
# Travel accessibility index calculator
# Uses OSGeo4W QGIS and QT libraries
# Uses Python 2.7
# Run using job_app.bat batch file to set up environment and run as Windows .pyw GUI
#
# Eddie Boyle 
# 01/02/16

#import libraries
from qgis.core import *
from qgis.gui import *
from qgis.utils import *
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.Qt import QShortcut, QFileDialog
from math import *
from sip import *
from functools import *

#set up QT/QGIS GUI application
app = QApplication([])
QgsApplication.setPrefixPath("OSGeo4W/apps/qgis", True)
QgsApplication.initQgis()

#create map canvas GUI widget with correct spatial parameters
canvas = QgsMapCanvas()
canvas.setCanvasColor(Qt.white)
canvas.mapRenderer().setDestinationCrs(QgsCoordinateReferenceSystem(27700, QgsCoordinateReferenceSystem.EpsgCrsId)) #BNG CRS
canvas.setMapUnits(QGis.Meters)

#function to load shapefile data layers from menubar with file browser dialog
def file_layer_show(layer_string):
    file_dialog = QFileDialog()
    file_title = 'Load data layer' 
    dialog_file_path = "TAA2Data"
    if layer_string == "basemap":
        file_name1 = QFileDialog.getOpenFileName(file_dialog, file_title, dialog_file_path)
        global layer1 #declare globally so it can be used in other functions
        layer1 = QgsRasterLayer(file_name1, "Basemap")
        # Set opacity
        layer1.renderer().setOpacity(0.5)
        layer1.setCrs(QgsCoordinateReferenceSystem(27700, QgsCoordinateReferenceSystem.EpsgCrsId)) #BNG CRS
        QgsMapLayerRegistry.instance().addMapLayer(layer1, True)
        canvas.setExtent(layer1.extent())
    elif layer_string == "jobs":
        file_name2 = QFileDialog.getOpenFileName(file_dialog, file_title, dialog_file_path)
        global layer2 #declare globally so it can be used in other functions
        layer2 = QgsVectorLayer(file_name2, "Jobs", "ogr")
        layer2.setCrs(QgsCoordinateReferenceSystem(27700, QgsCoordinateReferenceSystem.EpsgCrsId)) #BNG CRS
        QgsMapLayerRegistry.instance().addMapLayer(layer2, True)
        canvas.setExtent(layer2.extent())
    elif layer_string == "residents":
        file_name3 = QFileDialog.getOpenFileName(file_dialog, file_title, dialog_file_path)
        global layer3 #declare globally so it can be used in other functions
        layer3 = QgsVectorLayer(file_name3, "Residents", "ogr")
        layer3.setCrs(QgsCoordinateReferenceSystem(27700, QgsCoordinateReferenceSystem.EpsgCrsId)) #BNG CRS
        QgsMapLayerRegistry.instance().addMapLayer(layer3, True)
        canvas.setExtent(layer3.extent())
    elif layer_string == "clear":
        for layer in QgsMapLayerRegistry.instance().mapLayers().values():
            QgsMapLayerRegistry.instance().removeMapLayer(layer.id())
    canvas.refresh()

#function to calculate job accessibility and create temporary vector layer with data
jobAccessibilityLayer_count = 0
def job_acc_calc(tfc_text, dist_text):
    tfc=float(tfc_text) # user-supplied travel friction coefficient
    distance_limit=float(dist_text) # user-supplied distance cutoff in metres; NB index calculation uses km
    
    #only use features visible in map canvas - residents features are used
    request = QgsFeatureRequest()
    request.setFilterRect(canvas.extent())
    #count number of features being used
    request_count = 0
    for resident in layer3.getFeatures(request):
        request_count += 1
    
    #create temporary vector layer
    global jobAccessibilityLayer_count
    jobAccessibilityLayer_count += 1
    jobAccessibilityLayer_name = "job_accessibility_" + str(jobAccessibilityLayer_count)
    jobAccessibilityLayer = QgsVectorLayer("Point", jobAccessibilityLayer_name, "memory")
    pr = jobAccessibilityLayer.dataProvider()
    jobAccessibilityLayer.startEditing()
    pr.addAttributes([QgsField("OA11CD", QVariant.String), QgsField("Travel Accessibility Index", QVariant.Int)])
   
    index_count = 0
    #loop through all selected residents features and then for each feature loop through all job features, within a set distance
    for resident in layer3.getFeatures(request):
        r_attrs = resident.attributes()
        ta_index_sum = float(0)
        index_count += 1
        #job_total = 0
        for job in layer2.getFeatures():
            #make GUI responsive while loop runs
            app.processEvents()
            ta_index = float(0)
            j_attrs = job.attributes()
            dist = job.geometry().distance(resident.geometry()) #metres
            #job features within set distance limit
            if dist < distance_limit:
                #carry out accessibility index calculation
                no_jobs = j_attrs[2]
                dist_calc = dist/1000 #km
                ta_index = no_jobs * exp(-tfc * dist_calc) 
                ta_index_sum += ta_index
        # add result to temporary vector layer
        fet = QgsFeature()
        fet.setAttributes([r_attrs[0], ta_index_sum])
        fet.setGeometry(resident.geometry())
        pr.addFeatures([fet]) 
        index_string = "Index calculation count: " + str(index_count) + " of " + str(request_count)
        status_bar1.showMessage(index_string)
    
    #show temporary vector layer on canvas and legend
    jobAccessibilityLayer.commitChanges()
    jobAccessibilityLayer.setCrs(QgsCoordinateReferenceSystem(27700, QgsCoordinateReferenceSystem.EpsgCrsId)) #BNG CRS
    QgsMapLayerRegistry.instance().addMapLayer(jobAccessibilityLayer, True)
    canvas.setExtent(jobAccessibilityLayer.extent())
    canvas.refresh() 

#function to show dialog window to get user input for travel friction coefficient and distance cutoff
#and starting the index calculation
def tfc_dialog():
    #check that a residents data layer *and* a jobs data layer is loaded before doing the index calculation
    check = 1
    
    try:
        inputLayer = QgsMapLayerRegistry.instance().mapLayersByName("Residents")[0]
    except IndexError:
        QMessageBox.information(None, "Warning", "You need to load a residents data layer for this tool") 
        check = 0
    
    try:
        inputLayer = QgsMapLayerRegistry.instance().mapLayersByName("Jobs")[0]
    except IndexError:
        QMessageBox.information(None, "Warning", "You need to load a jobs data layer for this tool") 
        check = 0
        
    if check == 1:
        #create text input GUI widget 
        calc_input_field = QInputDialog()
        calc_title = "Parameters"
        tfc_label = "Enter a travel friction coefficient (0 - 1)"
        dist_label = "Enter a distance cutoff (metres)"
        tfc_mode = QLineEdit.Normal
        tfc_default = "0.15"   #0.15 default travel friction coefficient
        dist_default = "10000"   #10000 default distance cutoff (in metres)
        tfc_text, ok = QInputDialog.getText(calc_input_field, calc_title, tfc_label, tfc_mode, tfc_default)
        dist_text, ok = QInputDialog.getText(calc_input_field, calc_title, dist_label, tfc_mode, dist_default)
        job_acc_calc(tfc_text, dist_text)

#zoom in function
def zoom_in():
    canvas.setMapTool(toolZoomIn)
    
#zoom out function
def zoom_out():
    canvas.setMapTool(toolZoomOut)
    
#pan out function
def pan():
    canvas.setMapTool(toolPan)

#perform a clean exit
def exit_app():
    setdestroyonexit(False)
    qApp.quit()

#function to display help window
def help_w():
    help_string1 = "To run the job accessibility index calculator:\n1. Load a basemap layer using Layer>Load Basemap (optional; supplied file: 'manchester.tif')\n2. Load a jobs data layer using Layer>Load Jobs (required; supplied file 'Jobs.shp')\n3. Load a residents data layer using Layer>Load Residents (required; supplied file 'Residents.shp')\n4. Pan and zoom in or out using the 'Map' map canvas to show the extent of jobs point locations that you wish to calculate indexes for\n5. Start the calculation using Tools>Job Accessibility; the progress of the calculation can be seen in the status bar at bottom left of the application window\n6. When the calculation is finished, right click on the results layer ('job_accessibility_n') in the 'Layers' list to get the attributes or export the results as a shapefile"
    #create popup window to show help
    help_win = QDialog(window, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
    help_win.setWindowTitle("Help")
    #set initial window size and centre in desktop
    help_win.setGeometry(0, 0, 400, 300)
    resolution = QDesktopWidget().screenGeometry()
    help_win.move((resolution.width() / 2) - (help_win.frameSize().width() / 2), (resolution.height() / 2) - (help_win.frameSize().height() / 2))
    help_win_frame = QFrame(help_win)
    #use grid layout for widgets in help GUI window
    help_win_frame_layout = QGridLayout(help_win_frame)
    #create text GUI widget and add help text
    help_win_text = QTextEdit()
    help_win_text.setText(help_string1)
    help_win_text.setReadOnly(True)
    help_win_text.setMinimumWidth(378)
    help_win_text.setMinimumHeight(248)
    help_win_close_button = QDialogButtonBox(QDialogButtonBox.Close)
    help_win_close_button.rejected.connect(help_win.reject)
    help_win_frame_layout.addWidget(help_win_text, 0, 0)
    help_win_frame_layout.addWidget(help_win_close_button, 1, 0)
    help_win.show()
      
#function to display about window
def about_w():
    about_string1 = "UNIGIS MSc Spatial Databases & Programming module\nTAA2 PyQGIS Job Accessibility application\nAuthor: Eddie Boyle\nDate: April 2016\nBasemap: 1:250 000 Scale Colour Raster [TIFF geospatial data], Scale 1:250000, Tiles: sk,sj,se,sd, Updated: 24 April 2015, Ordnance Survey (GB), Using: EDINA Digimap Ordnance Survey Service, <http://digimap.edina.ac.uk>, Downloaded: 2016-01-21 14:05:06.671 "
    #create popup window to show about
    about_win = QDialog(window, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
    about_win.setWindowTitle("About")
    #set initial window size and centre in desktop
    about_win.setGeometry(0, 0, 400, 300)
    resolution = QDesktopWidget().screenGeometry()
    about_win.move((resolution.width() / 2) - (about_win.frameSize().width() / 2), (resolution.height() / 2) - (about_win.frameSize().height() / 2))
    about_win_frame = QFrame(about_win)
    #use grid layout for widgets in about GUI window
    about_win_frame_layout = QGridLayout(about_win_frame)
    #create text GUI widget and add about text
    about_win_text = QTextEdit()
    about_win_text.setText(about_string1)
    about_win_text.setReadOnly(True)
    about_win_text.setMinimumWidth(378)
    about_win_text.setMinimumHeight(248)
    about_win_close_button = QDialogButtonBox(QDialogButtonBox.Close)
    about_win_close_button.rejected.connect(about_win.reject)
    about_win_frame_layout.addWidget(about_win_text, 0, 0)
    about_win_frame_layout.addWidget(about_win_close_button, 1, 0)
    about_win.show()
    
#create main GUI window
window = QMainWindow()
window.setWindowTitle("PyQGIS Job Accessibility")
app.setActiveWindow(window)

#set initial window size and centre in desktop
window.setGeometry(0, 0, 850, 650)
resolution = QDesktopWidget().screenGeometry()
window.move((resolution.width() / 2) - (window.frameSize().width() / 2), (resolution.height() / 2) - (window.frameSize().height() / 2))
window_frame = QFrame(window)
window.setCentralWidget(window_frame)

#use grid layout for widgets in GUI window
frame_layout = QGridLayout(window_frame)

#create menubar GUI widget with keyboard shortcuts and icons
menu_bar = QMenuBar()
file_menu = menu_bar.addMenu("File")
exit_action = QAction(QIcon('icons/quit.png'), "Exit", window)
exit_action.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_Q))
window.connect(exit_action, SIGNAL("triggered()"), exit_app)
file_menu.addAction(exit_action)
layers_menu = menu_bar.addMenu("Layer")
file_layer1_action = QAction(QIcon('icons/layer-raster-add.png'), "Load Basemap (Raster)", window, checkable=False)
file_layer2_action = QAction(QIcon('icons/layer-vector-add.png'), "Load Jobs (Vector)", window, checkable=False)
file_layer3_action = QAction(QIcon('icons/layer-vector-add.png'), "Load Residents (Vector)", window, checkable=False)
file_layer4_action = QAction(QIcon('icons/layer-remove.png'), "Clear Layers", window, checkable=False)
file_layer1_action.triggered.connect(partial(file_layer_show, "basemap"))
file_layer2_action.triggered.connect(partial(file_layer_show, "jobs"))
file_layer3_action.triggered.connect(partial(file_layer_show, "residents"))
file_layer4_action.triggered.connect(partial(file_layer_show, "clear"))
layers_menu.addAction(file_layer1_action)
layers_menu.addAction(file_layer2_action)
layers_menu.addAction(file_layer3_action)
layers_menu.addAction(file_layer4_action)
tools_menu = menu_bar.addMenu("Tools")
zoom_in_action = QAction(QIcon('icons/zoom-in.png'), "Zoom In", window)
zoom_out_action = QAction(QIcon('icons/zoom-out.png'), "Zoom Out", window)
pan_action = QAction(QIcon('icons/pan.png'), "Pan", window)
acc_calc_action = QAction(QIcon('icons/execute.png'), "Job Accessibility", window)
zoom_in_action.triggered.connect(zoom_in)
zoom_out_action.triggered.connect(zoom_out)
pan_action.triggered.connect(pan)
acc_calc_action.triggered.connect(tfc_dialog)
tools_menu.addAction(zoom_in_action)
tools_menu.addAction(zoom_out_action)
tools_menu.addAction(pan_action)
tools_menu.addAction(acc_calc_action)
toolZoomIn = QgsMapToolZoom(canvas, False) # false = in
toolZoomIn.setAction(zoom_in_action)
toolZoomOut = QgsMapToolZoom(canvas, True) # true = out
toolZoomOut.setAction(zoom_out_action)
toolPan = QgsMapToolPan(canvas)
toolPan.setAction(pan_action)
help_menu = menu_bar.addMenu("Help")
help_action = QAction(QIcon('icons/help.png'), "Help", window)
about_action = QAction(QIcon('icons/about.png'), "About", window)
help_action.triggered.connect(help_w)
about_action.triggered.connect(about_w)
help_menu.addAction(help_action)
help_menu.addAction(about_action)
frame_layout.addWidget(menu_bar,0,0)

#create text label GUI widgets
label1 = QLabel()
label1.setText("Layers")
frame_layout.addWidget(label1,1,0)
label2 = QLabel()
label2.setText("Map")
frame_layout.addWidget(label2,1,1)

#create legend GUI widget
root = QgsProject.instance().layerTreeRoot()
bridge = QgsLayerTreeMapCanvasBridge(root, canvas)
model = QgsLayerTreeModel(root)
model.setFlag(QgsLayerTreeModel.AllowNodeReorder)
model.setFlag(QgsLayerTreeModel.AllowNodeRename)
model.setFlag(QgsLayerTreeModel.AllowNodeChangeVisibility)
model.setFlag(QgsLayerTreeModel.ShowLegend)
view = QgsLayerTreeView()
view.setModel(model)
        
#create class for right-click context menu for layers in legend
class MyMenuProvider(QgsLayerTreeViewMenuProvider):
    def __init__(self, view):
        QgsLayerTreeViewMenuProvider.__init__(self)
        self.view = view
    
    def createContextMenu(self):
        if not self.view.currentLayer():
            return None
        m = QMenu()
        if self.view.currentLayer().type() == QgsMapLayer.VectorLayer:
            m.addAction(QIcon('icons/attributes-display.png'), "Attributes", self.showAttributes)
            m.addAction(QIcon('icons/layer-export.png'), "Export as shapefile", self.exportLayer)
        m.addAction(QIcon('icons/zoom-layer.png'), "Zoom to layer", self.zoomLayer)
        m.addAction(QIcon('icons/layer-remove.png'), "Remove layer", self.removeLayer)
        return m

    def showAttributes(self):
        # retrieve every feature with its attributes
        features = self.view.currentLayer().getFeatures()
        title = self.view.currentLayer().name()
        display_string1 = "Fields:\t"
        for field in self.view.currentLayer().pendingFields():
            display_string1 += field.name() + "\t"
        feature_count = 0
        display_string3 = "Values:\t"
        for feature in features:
            feature_count += 1
            # fetch attributes
            attribute_list = feature.attributes()
            display_string3 += "\n"
            attr_count = 0
            for field in self.view.currentLayer().pendingFields():
                display_string3 += "\t"
                display_string3 += str(attribute_list[attr_count])
                attr_count += 1
        display_string2 = "No. of features: " + str(feature_count)
        #create popup window to show attributes
        attr = QDialog(window, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
        title_string = "Attributes of '" + title + "' layer"
        attr.setWindowTitle(title_string)
        #set initial window size and centre in desktop
        attr.setGeometry(0, 0, 600, 400)
        resolution = QDesktopWidget().screenGeometry()
        attr.move((resolution.width() / 2) - (attr.frameSize().width() / 2), (resolution.height() / 2) - (attr.frameSize().height() / 2))
        attr_frame = QFrame(attr)
        #use grid layout for widgets in attr GUI window
        attr_frame_layout = QGridLayout(attr_frame)
        #create text GUI widget and add attribute data
        attr_data = QTextEdit()
        attr_data.setText(display_string2)
        attr_data.append(display_string1)
        attr_data.append(display_string3)
        attr_data.setReadOnly(True)
        attr_data.setMinimumWidth(578)
        attr_data.setMinimumHeight(348)
        attr_table = QScrollArea()
        attr_table.setMinimumWidth(580)
        attr_table.setMinimumHeight(350)
        attr_close_button = QDialogButtonBox(QDialogButtonBox.Close)
        attr_close_button.rejected.connect(attr.reject)
        attr_table.setWidget(attr_data)
        attr_frame_layout.addWidget(attr_table, 0, 0)
        attr_frame_layout.addWidget(attr_close_button, 1, 0)
        attr.show()
     
    def removeLayer(self):
        #remove selected layer from the canvas and legend
        QgsMapLayerRegistry.instance().removeMapLayer(self.view.currentLayer().id())
        
    def zoomLayer(self):
        #zoom to the extent of the selected layer
        canvas.setExtent(self.view.currentLayer().extent())
        canvas.refresh() 
        
    def exportLayer(self):
        export_file_dialog = QFileDialog()
        export_file_title = 'Save data layer' 
        export_dialog_file_path = "TAA2Data"
        export_file_name = QFileDialog.getSaveFileName(export_file_dialog, export_file_title, export_dialog_file_path)
        shapefile = QgsVectorFileWriter.writeAsVectorFormat(self.view.currentLayer(), export_file_name, "utf-8" , None, "ESRI Shapefile")
        
provider = MyMenuProvider(view)
view.setMenuProvider(provider)

frame_layout.addWidget(view,2,0)

frame_layout.addWidget(canvas,2,1)

#create status bar GUI widget
status_bar1 = QStatusBar()
status_bar1.showMessage("Calculation progress")
frame_layout.addWidget(status_bar1, 3, 0)

#create status bar GUI widget
status_bar2 = QStatusBar()
status_bar2.showMessage(canvas.mapRenderer().destinationCrs().authid())
frame_layout.addWidget(status_bar2, 3, 1)

#show GUI window
window.show()

#make mouse pointer pan by default
pan()

#run QT GUI loop and exit application
exitcode = app.exec_()
QgsApplication.exitQgis()